# -*- coding: utf-8 -*-


def unpack_bcd(byte_array):
    half_byte_array = []
    for byte in byte_array:
        half_byte_array.append(byte >> 4 & 0x0F)
        half_byte_array.append(byte & 0x0F)
    return half_byte_array


def decode_bcd(half_byte_array):
    num = ""
    for byte in half_byte_array:
        byte &= 0b00001111
        assert byte < 10
        num += str(byte)
    return int(num)


def get_namespace(tree):
    """Create an artificial namespace to query xml.

    Try to extract the first namespace from the document. If there isn't any, assume a default namespace.
    This util is necessary since we do not know if prefix is used for the xml namespace or not.
    """
    keys = list(tree.nsmap.keys())
    if len(keys) > 0:
        key = keys[0]
        ns = tree.nsmap[key]
    else:
        ns = "http://ws.gematik.de/fa/vsdm/vsd/v5.2"

    return ns



# -*- coding: utf-8 -*-
import json
import zlib
from datetime import datetime
from json import JSONEncoder

from smartcard.Exceptions import NoCardException, CardConnectionException
from smartcard.System import readers

from healthcard.address import ResidenceAddress
from healthcard.exceptions import HealthCardException, HealthCardReadException
from healthcard.insurance import Insurance
from healthcard.patient import Patient
from healthcard.utils import unpack_bcd, decode_bcd

COMMANDS = {
    "SELECT_MF": [
        0x00,
        0xA4,
        0x04,
        0x0C,
        0x07,
        0xD2,
        0x76,
        0x00,
        0x01,
        0x44,
        0x80,
        0x00,
    ],
    "SELECT_HCA": [0x00, 0xA4, 0x04, 0x0C, 0x06, 0xD2, 0x76, 0x00, 0x00, 0x01, 0x02],
    "EF_GDO": [0x00, 0xB0, 0x82, 0x00, 0x00],
    "EF_VERSION_1": [0x00, 0xB2, 0x01, 0x84, 0x00],
    "EF_VERSION_2": [0x00, 0xB2, 0x02, 0x84, 0x00],
    "EF_VERSION_3": [0x00, 0xB2, 0x03, 0x84, 0x00],
    "SELECT_FILE_PD": [0x00, 0xB0, 0x81, 0x00, 0x02],
    "SELECT_FILE_VD": [0x00, 0xB0, 0x82, 0x00, 0x08],
    "RESET_CARD_TERMINAL": [0x20, 0x11, 0x00, 0x00, 0x00],
    "GET_CARD": [0x20, 0x12, 0x01, 0x00, 0x01, 0x05],
    "EJECT_CARD": [0x20, 0x15, 0x01, 0x00, 0x01, 0x01],
}

# SFID - Short file identifier. These are only valid in conjunction with the Parent-Directory (DF)
SFID = {
    "EF.ATR": 0x1D,  # in MF
    "EF.GDO": 0x02,  # in MF
    "EF.Version": 0x10,  # in MF
    "EF.StatusVD": 0x0C,  # in MF
    "EF.PD": 0x01,  # in HCA
    "EF.VD": 0x02,  # in HCA
}


class HealthCard(object):
    def __init__(self, generation, patient, insurance):
        self.patient = patient
        self.insurance = insurance
        self.version = generation

    def to_json(self):
        return HealthCardJSONEncoder().encode(self)

    def to_flattened_json(self):
        result = {
            "firstName": self.patient.first_name,
            "lastName": self.patient.last_name,
            "gender": self.patient.gender.lower(),
            "insuranceId": self.patient.insurant_id,
            "prefix": self.patient.prefix,
            "title": self.patient.title,
            "birthday": self.patient.birthdate.strftime("%d.%m.%Y"),
            "nameAddition": self.patient.name_addition,
        }
        if self.patient.residential_address:
            result.update(
                {
                    "city": self.patient.residential_address.city,
                    "zipCode": self.patient.residential_address.zip_code,
                    "country": self.patient.residential_address.country_code,
                    "street": self.patient.residential_address.street,
                    "streetNumber": self.patient.residential_address.street_number,
                    "addressAddition": self.patient.residential_address.address_addition,
                }
            )
        elif self.patient.postal_address:
            result.update(
                {
                    "city": self.patient.postal_address.city,
                    "zipCode": self.patient.postal_address.zip_code,
                    "country": self.patient.postal_address.country_code,
                    "mailbox": self.patient.postal_address.mailbox,
                }
            )
        return json.dumps(result)


class HealthCardReader(object):
    def __init__(self, index=0, cardservice=None):
        try:
            if cardservice:
                self.connection = cardservice.connection
            else:
                r = readers()
                if len(r) < 1:
                    raise HealthCardException("No reader found.")
                self.reader = r[index]
                self.connection = self.reader.createConnection()

            self.connection.connect()
        except (NoCardException, CardConnectionException):
            print("No card inserted for reader: {}".format(self.reader))

    def create_read_command(self, pos, length):
        bpos = [pos >> 8 & 0xFF, pos & 0xFF]
        return [0x00, 0xB0, bpos[0], bpos[1], length]

    def run_command(self, adpu):
        data, sw1, sw2 = self.connection.transmit(adpu)
        if (sw1, sw2) != (0x90, 0x00):
            if (sw1, sw2) == (0x62, 0x83):
                raise HealthCardException("File deactivated.")
            elif (sw1, sw2) == (0x6A, 0x82):
                raise HealthCardException("File not found.")
            elif (sw1, sw2) == (0x69, 0x00):
                raise HealthCardException("Command is not allowed.")
            elif (sw1, sw2) == (0x62, 0x82):
                raise HealthCardReadException("End Of File Reached Warning.")
            elif (sw1, sw2) == (0x62, 0x81):
                raise HealthCardReadException("Corrupted Data Warning.")
            elif (sw1, sw2) == (0x69, 0x86):
                raise HealthCardReadException("No current EF.")
            elif (sw1, sw2) == (0x69, 0x82):
                raise HealthCardReadException("Security Status not satisfied.")
            elif (sw1, sw2) == (0x69, 0x81):
                raise HealthCardReadException("Wrong File Type.")
            elif (sw1, sw2) == (0x6B, 0x00):
                raise HealthCardReadException("Offset too big.")
        return data

    def read_file(self, offset, length):
        data = []
        max_read = 0xFC
        pointer = offset
        while len(data) < length:
            bytes_left = length - len(data)
            readlen = bytes_left if bytes_left < max_read else max_read
            data_chunk = self.run_command(self.create_read_command(pointer, readlen))
            pointer += readlen
            data.extend(data_chunk)
        return data

    def get_version(self, adpu):
        data = self.run_command(adpu)
        hdata = unpack_bcd(data)
        version = "%i.%i.%i" % (
            decode_bcd(hdata[0:3]),
            decode_bcd(hdata[3:6]),
            decode_bcd(hdata[6:10]),
        )
        return version

    def get_card_generation(self):
        ef_version_1 = self.get_version(COMMANDS["EF_VERSION_1"])
        ef_version_2 = self.get_version(COMMANDS["EF_VERSION_2"])
        ef_version_3 = self.get_version(COMMANDS["EF_VERSION_3"])

        if (
            ef_version_1 == "3.0.0"
            and ef_version_2 == "3.0.0"
            and ef_version_3 == "3.0.2"
        ):
            generation = "G1"
        elif (
            ef_version_1 == "3.0.0"
            and ef_version_2 == "3.0.1"
            and ef_version_3 == "3.0.3"
        ):
            generation = "G1 plus"
        elif (
            ef_version_1 == "4.0.0"
            and ef_version_2 == "4.0.0"
            and ef_version_3 == "4.0.0"
        ):
            generation = "G2"
        else:
            generation = "unknown"
        return generation

    def get_health_card(self):

        # Select Masterfile (root)
        self.run_command(COMMANDS["SELECT_MF"])

        card_generation = self.get_card_generation()
        # Select Health Care Application
        self.run_command(COMMANDS["SELECT_HCA"])
        # Select file containing patient data
        self.run_command(COMMANDS["SELECT_FILE_PD"])

        # Create read command for the first two bytes of patient file.
        # Read first two byte of patient data. These contain the length of the PD file.
        data = self.run_command(self.create_read_command(0x00, 0x02))
        pd_length = (data[0] << 8) + data[1]
        # Since the two bytes are included themselves those two bytes are subtracted from the length.
        pd_length -= 0x02

        self.run_command(COMMANDS["SELECT_MF"])
        self.run_command(COMMANDS["SELECT_HCA"])
        # print('select pd')
        self.run_command(COMMANDS["SELECT_FILE_PD"])

        patient_data_compressed = self.read_file(0x02, pd_length)

        self.run_command(COMMANDS["SELECT_MF"])
        self.run_command(COMMANDS["SELECT_HCA"])
        # Select file containing insurance data
        self.run_command(COMMANDS["SELECT_FILE_VD"])

        # Read the first 8 byte of the EF.VD
        # The first two bytes contain the offset of the start of the unprotected insurance data.
        # The bytes 3 and 4 contain the offset of the end of the unprotected insurance data.
        # The bytes 5 and 6 contain the offset of the start of the protected insurance data.
        # The bytes 7 and 8 contain the offset of the end of the protected insurance data.
        data = self.run_command(self.create_read_command(0x00, 0x08))

        # Calculate the length of the unprotected insurance data.
        # Since we have zero based counting we need to subtract 1.
        vd_start = (data[0] << 8) + data[1]
        vd_end = (data[2] << 8) + data[3]
        vd_length = vd_end - (vd_start - 1)

        self.run_command(COMMANDS["SELECT_MF"])
        self.run_command(COMMANDS["SELECT_HCA"])
        self.run_command(COMMANDS["SELECT_FILE_VD"])

        insurance_data_compressed = self.read_file(vd_start, vd_length)

        patient_data_compressed.extend([0x00] * 16)
        patient_data_compressed = bytearray(patient_data_compressed)
        patient_data_compressed = bytes(patient_data_compressed)
        patient_data_xml = zlib.decompress(patient_data_compressed, 15 + 16)

        insurance_data_compressed.extend([0x00] * 16)
        insurance_data_compressed = bytearray(insurance_data_compressed)
        insurance_data_compressed = bytes(insurance_data_compressed)
        insurance_data_xml = zlib.decompress(insurance_data_compressed, 15 + 16)

        patient = Patient(patient_data_xml)
        insurance = Insurance(insurance_data_xml)

        self.connection.disconnect()

        return HealthCard(
            generation=card_generation, patient=patient, insurance=insurance
        )


class HealthCardJSONEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, (HealthCard, Insurance, Patient, ResidenceAddress)):
            return o.__dict__
        if type(o) is datetime:
            return o.strftime("%d.%m.%Y")
        if type(o) is bytes:
            try:
                return o.decode("iso-8859-15")
            except AttributeError:
                raise TypeError(repr(o) + " is not JSON serializable")
        raise TypeError(repr(o) + " is not JSON serializable")



# -*- coding: utf-8 -*-


class HealthCardException(Exception):
    pass


class HealthCardSelectException(HealthCardException):
    pass


class HealthCardReadException(HealthCardException):
    pass



"""Healthcard patient module.

This module allows to parse patient xml data from the eGK (Elektronische Gesundheitskarte) and store it into a python
object.

"""
# -*- coding: utf-8 -*-
from datetime import datetime
from lxml import etree
from pip._vendor.six import python_2_unicode_compatible

from healthcard.address import PostalAddress, ResidenceAddress
import sys

from healthcard.utils import get_namespace

sys.getdefaultencoding()


@python_2_unicode_compatible
class Patient(object):
    """Patient object which parses patient XML string into a python object."""

    def __init__(self, xml):
        self.xml = xml
        tree = etree.fromstring(self.xml)
        # xpath only works with a non-empty namespace
        # set namespace for vsd
        namsespaces = {}
        namsespaces["vsd"] = get_namespace(tree)

        text_list = tree.xpath("//vsd:Versicherten_ID//text()", namespaces=namsespaces)
        self.insurant_id = text_list[0] if len(text_list) != 0 else ""
        text_list = tree.xpath("//vsd:Geburtsdatum//text()", namespaces=namsespaces)
        self.birthdate = text_list[0] if len(text_list) != 0 else ""
        self.birthdate = datetime.strptime(self.birthdate, "%Y%m%d")
        text_list = tree.xpath("//vsd:Vorname//text()", namespaces=namsespaces)
        self.first_name = text_list[0] if len(text_list) != 0 else ""
        text_list = tree.xpath("//vsd:Nachname//text()", namespaces=namsespaces)
        self.last_name = text_list[0] if len(text_list) != 0 else ""
        text_list = tree.xpath("//vsd:Geschlecht//text()", namespaces=namsespaces)
        self.gender = text_list[0] if len(text_list) != 0 else ""
        text_list = tree.xpath("//vsd:Vorsatzwort//text()", namespaces=namsespaces)
        self.prefix = text_list[0] if len(text_list) != 0 else ""
        text_list = tree.xpath("//vsd:Namenszusatz//text()", namespaces=namsespaces)
        self.name_addition = text_list[0] if len(text_list) != 0 else ""
        text_list = tree.xpath("//vsd:Titel//text()", namespaces=namsespaces)
        self.title = text_list[0] if len(text_list) != 0 else ""

        postal_address_node = tree.xpath(
            "//vsd:PostfachAdresse", namespaces=namsespaces
        )
        if postal_address_node:
            self.postal_address = PostalAddress(postal_address_node[0], namsespaces)
        residential_address_node = tree.xpath(
            "//vsd:StrassenAdresse", namespaces=namsespaces
        )
        if residential_address_node:
            self.residential_address = ResidenceAddress(
                residential_address_node[0], namsespaces
            )

    @property
    def address(self):
        return (
            self.postal_address
            if getattr(self, "postal_address", None)
            else self.residential_address
        )

    def to_json(self):
        from healthcard.reader import HealthCardJSONEncoder

        return HealthCardJSONEncoder().encode(self)

    def __str__(self):
        return u"{} {}".format(self.first_name, self.last_name)



# -*- coding: utf-8 -*-
"""Address module to parse lxml's etree._Element objects for address data.

There are 2 different formats in which the XML could be read from the eGK (Elektronische Gesundheitskarte):
VSD 5.1.x and VSD 5.2.x

There is a good source which explains the main difference between both formats:
ftp://ftp.kbv.de/ita-update/Abrechnung/KBV_ITA_VGEX_Mapping_KVK.pdf (page 9)
"""
from pip._vendor.six import python_2_unicode_compatible


@python_2_unicode_compatible
class Address(object):
    def __init__(self, root, nsmap):
        text_list = root.xpath("//vsd:Ort//text()", namespaces=nsmap)
        self.city = text_list[0] if len(text_list) != 0 else ""
        text_list = root.xpath("//vsd:Postleitzahl//text()", namespaces=nsmap)
        self.zip_code = text_list[0] if len(text_list) != 0 else ""
        text_list = root.xpath("//vsd:Wohnsitzlaendercode//text()", namespaces=nsmap)
        self.country_code = text_list[0] if len(text_list) != 0 else ""

    def __str__(self):
        return u"{}-{} {}".format(self.country_code, self.zip_code, self.city)


@python_2_unicode_compatible
class PostalAddress(Address):
    def __init__(self, root, nsmap):
        """Parses an etree._Element for postal address data.

        Since there are two different version for the format of VSD xml data (5.1.x and 5.2.x) there are different tags
        needed to access the same property. If VSD 5.1.x is not found we try to access the VSD 5.2.x version of a tag.
        """
        text_list = root.xpath("//vsd:Ort//text()", namespaces=nsmap)
        self.city = text_list[0] if len(text_list) != 0 else ""
        text_list = root.xpath("//vsd:Postleitzahl//text()", namespaces=nsmap)
        self.zip_code = text_list[0] if len(text_list) != 0 else ""
        text_list = root.xpath("//vsd:Postfach//text()", namespaces=nsmap)
        self.mailbox = text_list[0] if len(text_list) != 0 else ""
        super(PostalAddress, self).__init__(root, nsmap)

    def __str__(self):
        return u"{}-{} {} - Mailbox {}".format(
            self.country_code, self.zip_code, self.city, self.mailbox
        )


@python_2_unicode_compatible
class ResidenceAddress(Address):
    """Parses an etree._Element object for residential address data."""

    def __init__(self, root, nsmap):
        text_list = root.xpath("//vsd:Strasse//text()", namespaces=nsmap)
        self.street = text_list[0] if len(text_list) != 0 else ""
        text_list = root.xpath("//vsd:Hausnummer//text()", namespaces=nsmap)
        self.street_number = text_list[0] if len(text_list) != 0 else ""
        text_list = root.xpath("//vsd:Anschriftenzusatz//text()", namespaces=nsmap)
        self.address_addition = text_list[0] if len(text_list) != 0 else ""
        super(ResidenceAddress, self).__init__(root, nsmap)

    def __str__(self):
        return u"{}-{} {} - {} {}".format(
            self.country_code,
            self.zip_code,
            self.city,
            self.street,
            self.street_number,
            self.address_addition,
        )



# -*- coding: utf-8 -*-
import pathlib
from setuptools import setup
from healthcard import __version__

DIR = pathlib.Path(__file__).parent

README = (DIR / "README.rst").read_text()


setup(
    name="python-healthcard",
    version=__version__,
    description="A module to read german health insurance cards with python",
    url="https://github.com/Blueshoe/python-healthcard/",
    author="Blueshoe",
    author_email="robert@blueshoe.de",
    license="MIT",
    long_description=README,
    long_description_content_type="text/x-rst",
    packages=["healthcard"],
    install_requires=["pyscard==1.9.6", "lxml==4.6.3"],
    zip_safe=False,
)



"""Healthcard insurance module.

This module allows to parse insurance data XML and store insurance data into a python object.
"""
# -*- coding: utf-8 -*-
from lxml import etree

from .utils import get_namespace


class Insurance(object):
    def __init__(self, xml):
        self.xml = xml.decode("iso-8859-15")
        tree = etree.fromstring(xml)
        namsespaces = {}
        namsespaces["vsd"] = get_namespace(tree)

        cost_node = tree.xpath("//vsd:Kostentraeger", namespaces=namsespaces)[0]
        text_list = cost_node.xpath("vsd:Name//text()", namespaces=namsespaces)
        self.insurance_name = text_list[0] if len(text_list) != 0 else ""
